#include "geodesy.h"
#include <math.h>

inline double to_rad(double angle)
{
	return 2.0*M_PI*angle/360.0;
}

inline double to_deg(double angle)
{
	return 360.0*angle/(2.0*M_PI);
}

inline double decimal_deg(int degrees, int minutes, double seconds)
{
	return degrees + minutes/60.0 + seconds/3600.0;
}

double shortest_angle (double from_angle, double to_angle)
{
	double x = to_angle - from_angle;
	if (x < -M_PI)
		x += 2.0*M_PI;
	else if (x > M_PI)
		x -= 2.0*M_PI;
	return x;
}

void vincenty_inverse (double lon1, double lat1, double lon2, double lat2, double h,
	double *s, double *a1, double *a2)
{
	int iterLimit;
	double a,b,f,L,U1,U2,sinU1,cosU1,sinU2,cosU2,lmbda,lmbdaP;
	double sinLambda,cosLambda,sinSigma,cosSigma,sigma,sinAlpha,cosSqAlpha;
	double cos2SigmaM,C,uSq,A,B,deltaSigma,tmp;

	a = 6378137 + h;
	b = 6356752.3142 + h;
	f = 1/298.257223563;
	L = lon2-lon1;
	U1 = atan((1.0-f)*tan(lat1));
	U2 = atan((1.0-f)*tan(lat2));
	sinU1 = sin(U1);
	cosU1 = cos(U1);
	sinU2 = sin(U2);
	cosU2 = cos(U2);
  
	lmbda = L;
	lmbdaP = 2.0*M_PI;
	iterLimit = 20;
	while (abs(lmbda-lmbdaP) > 1e-14 && iterLimit>0) {
		sinLambda = sin(lmbda);
		cosLambda = cos(lmbda);
		sinSigma = sqrt((cosU2*sinLambda)*(cosU2*sinLambda)+(cosU1*sinU2-sinU1*cosU2*cosLambda)*(cosU1*sinU2-sinU1*cosU2*cosLambda));
		if (sinSigma==0) {
			*s = 0.0;
			*a1 = *a2 = 0.0;
			return; // co-incident points
		}
		cosSigma = sinU1*sinU2+cosU1*cosU2*cosLambda;
		sigma = atan2(sinSigma,cosSigma);
		sinAlpha = cosU1*cosU2*sinLambda/sinSigma;
		cosSqAlpha = 1.0-sinAlpha*sinAlpha;
		if (cosSqAlpha == 0.0)
			cos2SigmaM = 0.0; // equatorial line: cosSqAlpha=0
		else
			cos2SigmaM = cosSigma-2*sinU1*sinU2/cosSqAlpha;
		C = f/16.0*cosSqAlpha*(4.0+f*(4.0-3.0*cosSqAlpha));
		lmbdaP = lmbda;
		lmbda = L+(1.0-C)*f*sinAlpha*(sigma+C*sinSigma*(cos2SigmaM+C*cosSigma*(-1.0+2.0*cos2SigmaM*cos2SigmaM)));
		iterLimit-=1;
	}
	if (iterLimit == 0) {
		*s = -1.0;
		*a1 = *a2 = 0.0;
		return; // Failed to converge
	}

	uSq = cosSqAlpha*(a*a-b*b)/(b*b);
	A = 1.0 + uSq/16384.0*(4096.0+uSq*(-768.0+uSq*(320.0-175.0*uSq)));
	B = uSq/1024.0 * (256.0+uSq*(-128.0+uSq*(74.0-47.0*uSq)));
	deltaSigma = B*sinSigma*(cos2SigmaM+B/4.0*(cosSigma*(-1.0+2.0*cos2SigmaM*cos2SigmaM)-
		B/6.0*cos2SigmaM*(-3.0+4.0*sinSigma*sinSigma)*(-3.0+4.0*cos2SigmaM*cos2SigmaM)));
	*s = b*A*(sigma-deltaSigma);

	*a1 = atan2(cosU2*sin(lmbda), cosU1*sinU2 - sinU1*cosU2*cos(lmbda));
	if (*a1 < 0.0)
		*a1 += 2.0*M_PI;
	*a2 = atan2(cosU1*sin(lmbda), -sinU1*cosU2 + cosU1*sinU2*cos(lmbda));
	if (*a2 < 0.0)
		*a2 += 2*M_PI;
}

void vincenty_direct (double lon1, double lat1, double brng, double dist, double h,
	double *lon2, double *lat2)
{
	double a,b,f,s,alpha1,sinAlpha1,cosAlpha1,tanU1,cosU1,sinU1,sigma1;
	double sinAlpha,cosSqAlpha,uSq,A,B,sigma,sigmaP,cos2SigmaM,sinSigma;
	double cosSigma,deltaSigma,tmp,lmbda,C,L;

	a = 6378137.0 + h;
	b = 6356752.3142 + h;
	f = 1.0/298.257223563;
	s = dist;
	alpha1 = brng;
	sinAlpha1 = sin(alpha1);
	cosAlpha1 = cos(alpha1);
  
	tanU1 = (1.0-f)*tan(lat1);
	cosU1 = 1.0 / sqrt((1.0 + tanU1*tanU1));
	sinU1 = tanU1*cosU1;
	sigma1 = atan2(tanU1, cosAlpha1);
	sinAlpha = cosU1 * sinAlpha1;
	cosSqAlpha = 1.0 - sinAlpha*sinAlpha;
	uSq = cosSqAlpha * (a*a - b*b) / (b*b);
	A = 1.0 + uSq/16384.0*(4096.0+uSq*(-768.0+uSq*(320.0-175.0*uSq)));
	B = uSq/1024.0 * (256.0+uSq*(-128.0+uSq*(74.0-47.0*uSq)));
  
	sigma = s / (b*A);
	sigmaP = 2*M_PI;
	while (abs(sigma-sigmaP) > 1e-7) { // was 1e-14
		cos2SigmaM = cos(2.0*sigma1 + sigma);
		sinSigma = sin(sigma);
		cosSigma = cos(sigma);
		deltaSigma = B*sinSigma*(cos2SigmaM+B/4.0*(cosSigma*(-1.0+2.0*cos2SigmaM*cos2SigmaM)-
			B/6.0*cos2SigmaM*(-3.0+4.0*sinSigma*sinSigma)*(-3.0+4.0*cos2SigmaM*cos2SigmaM)));
		sigmaP = sigma;
		sigma = s / (b*A) + deltaSigma;
	}
	tmp = sinU1*sinSigma - cosU1*cosSigma*cosAlpha1;
	*lat2 = atan2(sinU1*cosSigma + cosU1*sinSigma*cosAlpha1, (1.0-f)*sqrt(sinAlpha*sinAlpha + tmp*tmp));
	lmbda = atan2(sinSigma*sinAlpha1, cosU1*cosSigma - sinU1*sinSigma*cosAlpha1);
	C = f/16.0*cosSqAlpha*(4.0+f*(4.0-3.0*cosSqAlpha));
	L = lmbda - (1.0-C) * f * sinAlpha * (sigma + C*sinSigma*(cos2SigmaM+C*cosSigma*(-1.0+2.0*cos2SigmaM*cos2SigmaM)));
	*lon2 = lon1 + L;
}

void xyz2llh (double X,double Y,double Z, double *lon, double *lat, double *h)
{
	double a = 6378137.0;
	double f = 1.0/298.257223563;
	double b,e2,ep2,r2,r,E2,F,G,c,s,P,Q,ro,tmp,U,V,zo;

	b = a*(1.0-f);
 
	e2 = 2.0*f-f*f;
	ep2 = f*(2.0-f)/((1.0-f)*(1.0-f));
 
	r2 = X*X+Y*Y;
	r = sqrt(r2);
	E2 = a*a - b*b;
	F = 54.0*b*b*Z*Z;
	G = r2 + (1.0-e2)*Z*Z - e2*E2;
	c = (e2*e2*F*r2)/(G*G*G);
	s = pow(1.0 + c + sqrt(c*c + 2.0*c),(1.0/3.0));
	P = F/(3.0*(s+1.0/s+1.0)*(s+1.0/s+1.0)*G*G);
	Q = sqrt(1.0+2.0*e2*e2*P);
	ro = -(e2*P*r)/(1.0+Q) + sqrt((a*a/2.0)*(1.0+1.0/Q) - ((1.0-e2)*P*Z*Z)/(Q*(1.0+Q)) - P*r2/2.0);
	tmp = (r - e2*ro)*(r - e2*ro);
	U = sqrt( tmp + Z*Z );
	V = sqrt( tmp + (1.0-e2)*Z*Z );
	zo = (b*b*Z)/(a*V);
 
	*h = U*( 1.0 - b*b/(a*V));
	*lat = atan( (Z + ep2*zo)/r );
	*lon = atan2(Y,X);
}

void llh2xyz (double lon, double lat, double h, double *x, double *y, double *z)
{
	// mathworks aerospace blockset
	double R = 6378137.0;
	double f = 1.0/298.257223563;
	double lats, rs;

	lats = atan(pow(1.0-f,2.0)*tan(lat));
	rs = sqrt(pow(R,2.0)/(1.0 + (1.0/pow((1.0-f),2.0) - 1.0)*sin(lats)*sin(lats)));
	*x = rs*cos(lats)*cos(lon) + h*cos(lat)*cos(lon);
	*y = rs*cos(lats)*sin(lon) + h*cos(lat)*sin(lon);
	*z = rs*sin(lats) + h*sin(lat);
}

void genmsg_data (double lon, double lat, double height, double course, double speed,
	double *x1, double *y1, double *z1,
	double *dx, double *dy, double *dz)
{
	double lon2, lat2, x2, y2, z2;

	llh2xyz(lon, lat, height, x1, y1, z1);
	vincenty_direct (lon,lat,course,speed,height, &lon2, &lat2);
	llh2xyz(lon2, lat2, height, &x2, &y2, &z2);
	*dx = x2-*x1;
	*dy = y2-*y1;
	*dz = z2-*z1;
}
