#include <geos_c.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>

#include "geodesy.h"

#define MIN(a,b) ((a<b)?(a):(b))
#define MAX(a,b) ((a>b)?(a):(b))

typedef struct area {
	struct area *next;
	int id;
	GEOSGeometry *geom;
} area_t;

area_t *areas = NULL;

area_t *area_insert(area_t *area_list, area_t *area)
{
	static int id=1;
	area->id = id;
	printf("Added area %d\n",area->id);
	id++;
	area->next = area_list;
	return area;
}

area_t *area_remove(area_t *area_list, area_t *area)
{
	area_t *tmp;
	printf ("Removed area %d\n", area->id);
	if (area_list == area)
		return area_list->next;
	tmp = area_list;
	while(tmp->next != NULL) {
		if (tmp->next == area)
			tmp->next = area->next;
		else
			tmp = tmp->next;
	}
	return area_list;
}

void notice(const char *fmt, ...)
{
        va_list ap;

        fprintf( stdout, "NOTICE: ");
        va_start (ap, fmt);
        vfprintf( stdout, fmt, ap);
        va_end(ap);
        fprintf( stdout, "\n" );
}

void log_and_exit(const char *fmt, ...)
{
        va_list ap;

        fprintf( stdout, "ERROR: ");
        va_start (ap, fmt);
        vfprintf( stdout, fmt, ap);
        va_end(ap);
        fprintf( stdout, "\n" );
        exit(1);
}

area_t *radar_coverage(double lng, double lat, double rng)
{
	int i;
	double lng2,lat2;
	GEOSCoordSequence *cs;
	GEOSGeometry *lr, *poly;
	area_t *area;

	area = malloc(sizeof(area_t));
	area->next = NULL;

	cs = GEOSCoordSeq_create(361,2);
	for(i=0;i<360;i++) {
		double brg = 2.0*M_PI*(double)i/360.0;
		vincenty_direct (lng, lat, brg, rng, 0.0, &lng2, &lat2);
		GEOSCoordSeq_setX(cs,i,lng2);
		GEOSCoordSeq_setY(cs,i,lat2);
	}
	GEOSCoordSeq_getX(cs,0,&lng2);
	GEOSCoordSeq_setX(cs,360,lng2);
	GEOSCoordSeq_getY(cs,0,&lat2);
	GEOSCoordSeq_setY(cs,360,lat2);

	lr = GEOSGeom_createLinearRing(cs);
	poly = GEOSGeom_createPolygon(lr, NULL, 0);

	area->geom = poly;

	return area;
}

void collection_contents(GEOSGeometry *g)
{
	int i, num;
	num = GEOSGetNumGeometries(g);
	printf("--%d subgeometries\n",num);
	printf("--isEmpty = %d\n",GEOSisEmpty(g));
	for (i=0;i<GEOSGetNumGeometries(g);i++) {
		double area;
		GEOSGeometry const *sub;
		sub = GEOSGetGeometryN(g,i);
		GEOSArea(sub,&area);
		printf("--subgeometry typeid = %d, area = %f\n", GEOSGeomTypeId(sub), area);
	}
}

void check(char *str, GEOSGeometry *g1,GEOSGeometry *g2, GEOSGeometry *test)
{
	int type = GEOSGeomTypeId(test);
	if (type != 3 && type != 6) {
		GEOSGeometry *tmp;
		printf("Un-expected result\n");
		printf ("%s\n", str);
		printf("Result TypeId = %d\n", GEOSGeomTypeId(test));
		printf("Result isEmpty = %d\n", GEOSisEmpty(test));
		printf("Input g1 TypeId = %d\n", GEOSGeomTypeId(g1));
		printf("Input g2 TypeId = %d\n", GEOSGeomTypeId(g2));
		printf("g1 disjoint g2 = %d\n", GEOSDisjoint(g1,g2));
		printf("g1 touches g2 = %d\n", GEOSTouches(g1,g2));
		printf("g1 overlaps g2 = %d\n", GEOSOverlaps(g1,g2));
		if (GEOSGeomTypeId(test) == 7)
			collection_contents(test);
		exit(-1);
	}
}

void add(GEOSGeometry *geom)
// Expects only type 3 and 6
// Takes ownership
{
	area_t *tmp;
	if (GEOSGeomTypeId(geom) == 3) {
		tmp = malloc(sizeof(area_t));
		tmp->geom = geom;
		areas = area_insert(areas, tmp);
	} else if (GEOSGeomTypeId(geom) == 6) {
		double a;
		int i, num;
		printf("multi polygon result\n");
		GEOSArea(geom,&a);
		printf("area of multipolygon = %f\n", a);
		num = GEOSGetNumGeometries(geom);
		for(i=0;i<num;i++) {
			GEOSGeometry const *sub_geom;
			sub_geom = GEOSGetGeometryN(geom,i);
			GEOSArea(sub_geom,&a);
			if (GEOSGeomTypeId(sub_geom) != 3) {
				printf("multipolygon sub geom is not a polygon\n");
				exit(-1);
			}
			printf("area of sub polygon = %f\n", a);
			if (a > 0.0) {
				tmp = malloc(sizeof(area_t));
				tmp->geom = GEOSGeom_clone(sub_geom);
				areas = area_insert(areas, tmp);
			} else
				printf("dropping zero area polygon from multipolygon\n");
		}
		GEOSGeom_destroy(geom);
	}
}

int divide(area_t *a1, area_t *a2, area_t *new_areas)
{
	int intersection_exists = 0;
	int diff_1_exists = 0;
	int diff_2_exists = 0;
	GEOSGeometry *gi,*diff, *tmp1, *tmp2;

	tmp1 = GEOSIntersection(a1->geom, a2->geom);
	tmp2 = GEOSSimplify(tmp1,1e-9);
	GEOSGeom_destroy(tmp1);
	gi = GEOSBuffer(tmp2,0.0,0);
	GEOSGeom_destroy(tmp2);

	check("intersection", a1->geom,a2->geom,gi);
	if (!GEOSisEmpty(gi)) {
		double a;
		GEOSArea(gi,&a);
		printf("Intersection is not empty, type = %d, Area = %0.20f\n",GEOSGeomTypeId(gi),a);
		add(gi);
		intersection_exists = 1;
	} else {
		printf("Intersection is empty, returning early\n");
		return 0;
	}

	tmp1 = GEOSDifference(a1->geom,gi);
	tmp2 = GEOSSimplify(tmp1,1e-9);
	GEOSGeom_destroy(tmp1);
	diff = GEOSBuffer(tmp2,0.0,0);
	if (!GEOSisEmpty(diff)) {
		double a;
		GEOSArea(gi,&a);
		printf("1st diff is not empty, type = %d, Area = %0.20f\n",GEOSGeomTypeId(diff),a);
		check("1st diff", a1->geom, gi, diff);
		diff_1_exists = 1;
		add(diff);
	} else
		printf("1st diff is empty\n");

	tmp1 = GEOSDifference(a2->geom,gi);
	tmp2 = GEOSSimplify(tmp1,1e-9);
	GEOSGeom_destroy(tmp1);
	diff = GEOSBuffer(tmp2,0.0,0);
	if (!GEOSisEmpty(diff)) {
		double a;
		GEOSArea(gi,&a);
		printf("2nd diff is not empty, type = %d, Area = %0.20f\n",GEOSGeomTypeId(diff),a);
		check("2nd diff", a2->geom, gi, diff);
		diff_2_exists = 1;
		add(diff);
	} else
		printf("2nd diff is empty\n");

	if (intersection_exists==1 || diff_1_exists==1 || diff_2_exists==1) {
		areas = area_remove(areas, a1);
		areas = area_remove(areas, a2);
		printf("divide done\n");
		return 1;
	}

	printf("divide done (NOTHING TO DIVIDE)\n");
	return 0;
}

int split_areas()
{
	area_t *a1,*a2;

	a1 = areas;
	while(a1 != NULL) {
		a2 = areas;
		while(a2 != NULL) {
			if (a1 != a2)
				if (GEOSIntersects(a1->geom,a2->geom)) {
					printf("%d and %d intersect\n",a1->id,a2->id);
					if (divide(a1,a2,areas))
						return 1;
					else
						printf("divide areas did nothing\n");
				}
			a2 = a2->next;
		}
		a1 = a1-> next;
	}
	return 0;
}

int main(int argc, char *argv[])
{
	int i;

	initGEOS(notice, log_and_exit);

	// Test 3 overlapping circles in a line, resulting in multipolygons ( PASSES )
	areas = area_insert(areas, radar_coverage(2.5, -0.5, 250000.0));
	areas = area_insert(areas, radar_coverage(2.515, -0.5, 250000.0));
	areas = area_insert(areas, radar_coverage(2.53, -0.5, 250000.0));
	
	// Test two identical areas ( PASSES )
	//areas = area_insert(areas, radar_coverage(2.5, -0.5, 250000.0));
	//areas = area_insert(areas, radar_coverage(2.5, -0.5, 250000.0));

	// Test two coincident radars creating a hole ( PASSES )
	//areas = area_insert(areas, radar_coverage(2.5, -0.5, 250000.0));
	//areas = area_insert(areas, radar_coverage(2.5, -0.5, 150000.0));

	// Stress test radars crammed together (FAILS)
	//for(i=0;i<5;i++) // Fails side location conflict
	//	areas = area_insert(areas, radar_coverage(2.5+drand48()/10.0, -0.5+drand48()/10.0, 250000.0));

	//for(i=0;i<5;i++) // Fails stuck in endless loop
	//	areas = area_insert(areas, radar_coverage(2.5+drand48()/100.0, -0.5+drand48()/100.0, 250000.0));

	//for(i=0;i<5;i++) // Fails non-noded intersection
	//	areas = area_insert(areas, radar_coverage(2.5+drand48()/1000.0, -0.5+drand48()/1000.0, 250000.0));
	
	//for(i=0;i<30;i++) // Fails  no outgoing dirEdge found
	//	areas = area_insert(areas, radar_coverage(2.5+drand48()/2.0, -0.5+drand48()/1000.0, 250000.0));

	while(split_areas());

	finishGEOS();
	return 0;
}
